@@ -1,6 +1,6 @@
 /*
     dtach - A simple program that emulates the detach feature of screen.
-    Copyright (C) 2004-2008 Ned T. Crigler
+    Copyright (C) 2004-2016 Ned T. Crigler
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -52,6 +52,12 @@ connect_socket(char *name)
 	int s;
 	struct sockaddr_un sockun;
 
+	if (strlen(name) > sizeof(sockun.sun_path) - 1)
+	{
+		errno = ENAMETOOLONG;
+		return -1;
+	}
+
 	s = socket(PF_UNIX, SOCK_STREAM, 0);
 	if (s < 0)
 		return -1;
@@ -150,6 +156,28 @@ attach_main(int noerror)
 	/* Attempt to open the socket. Don't display an error if noerror is 
 	** set. */
 	s = connect_socket(sockname);
+	if (s < 0 && errno == ENAMETOOLONG)
+	{
+		char *slash = strrchr(sockname, '/');
+
+		/* Try to shorten the socket's path name by using chdir. */
+		if (slash)
+		{
+			int dirfd = open(".", O_RDONLY);
+
+			if (dirfd >= 0)
+			{
+				*slash = '\0';
+				if (chdir(sockname) >= 0)
+				{
+					s = connect_socket(slash + 1);
+					fchdir(dirfd);
+				}
+				*slash = '/';
+				close(dirfd);
+			}
+		}
+	}
 	if (s < 0)
 	{
 		if (!noerror)
@@ -190,6 +218,7 @@ attach_main(int noerror)
 	write(1, "\33[H\33[J", 6);
 
 	/* Tell the master that we want to attach. */
+	memset(&pkt, 0, sizeof(struct packet));
 	pkt.type = MSG_ATTACH;
 	write(s, &pkt, sizeof(struct packet));
 
@@ -217,7 +246,7 @@ attach_main(int noerror)
 		/* Pty activity */
 		if (n > 0 && FD_ISSET(s, &readfds))
 		{
-			int len = read(s, buf, sizeof(buf));
+			ssize_t len = read(s, buf, sizeof(buf));
 
 			if (len == 0)
 			{
@@ -237,12 +266,16 @@ attach_main(int noerror)
 		/* stdin activity */
 		if (n > 0 && FD_ISSET(0, &readfds))
 		{
+			ssize_t len;
+
 			pkt.type = MSG_PUSH;
 			memset(pkt.u.buf, 0, sizeof(pkt.u.buf));
-			pkt.len = read(0, pkt.u.buf, sizeof(pkt.u.buf));
+			len = read(0, pkt.u.buf, sizeof(pkt.u.buf));
 
-			if (pkt.len <= 0)
+			if (len <= 0)
 				exit(1);
+
+			pkt.len = len;
 			process_kbd(s, &pkt);
 			n--;
 		}
@@ -259,3 +292,70 @@ attach_main(int noerror)
 	}
 	return 0;
 }
+
+int
+push_main()
+{
+	struct packet pkt;
+	int s;
+
+	/* Attempt to open the socket. */
+	s = connect_socket(sockname);
+	if (s < 0 && errno == ENAMETOOLONG)
+	{
+		char *slash = strrchr(sockname, '/');
+
+		/* Try to shorten the socket's path name by using chdir. */
+		if (slash)
+		{
+			int dirfd = open(".", O_RDONLY);
+
+			if (dirfd >= 0)
+			{
+				*slash = '\0';
+				if (chdir(sockname) >= 0)
+				{
+					s = connect_socket(slash + 1);
+					fchdir(dirfd);
+				}
+				*slash = '/';
+				close(dirfd);
+			}
+		}
+	}
+	if (s < 0)
+	{
+		printf("%s: %s: %s\n", progname, sockname, strerror(errno));
+		return 1;
+	}
+
+	/* Set some signals. */
+	signal(SIGPIPE, SIG_IGN);
+
+	/* Push the contents of standard input to the socket. */
+	pkt.type = MSG_PUSH;
+	for (;;)
+	{
+		ssize_t len;
+
+		memset(pkt.u.buf, 0, sizeof(pkt.u.buf));
+		len = read(0, pkt.u.buf, sizeof(pkt.u.buf));
+
+		if (len == 0)
+			return 0;
+		else if (len < 0)
+		{
+			printf("%s: %s: %s\n", progname, sockname,
+			       strerror(errno));
+			return 1;
+		}
+
+		pkt.len = len;
+		if (write(s, &pkt, sizeof(struct packet)) < 0)
+		{
+			printf("%s: %s: %s\n", progname, sockname,
+			       strerror(errno));
+			return 1;
+		}
+	}
+}
