@@ -1,6 +1,6 @@
 /*
     dtach - A simple program that emulates the detach feature of screen.
-    Copyright (C) 2001, 2004-2008 Ned T. Crigler
+    Copyright (C) 2001, 2004-2016 Ned T. Crigler
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -13,8 +13,7 @@
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
-    along with this program; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 #ifndef dtach_h
 #define dtach_h
@@ -128,7 +127,8 @@ struct packet
 #define EOS "\033[999H"
 
 int attach_main(int noerror);
-int master_main(char **argv, int waitattach);
+int master_main(char **argv, int waitattach, int dontfork);
+int push_main(void);
 
 #ifdef sun
 #define BROKEN_MASTER
