@@ -1,6 +1,6 @@
 /*
     dtach - A simple program that emulates the detach feature of screen.
-    Copyright (C) 2004-2008 Ned T. Crigler
+    Copyright (C) 2004-2016 Ned T. Crigler
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -185,6 +185,12 @@ create_socket(char *name)
 	int s;
 	struct sockaddr_un sockun;
 
+	if (strlen(name) > sizeof(sockun.sun_path) - 1)
+	{
+		errno = ENAMETOOLONG;
+		return -1;
+	}
+
 	s = socket(PF_UNIX, SOCK_STREAM, 0);
 	if (s < 0)
 		return -1;
@@ -214,13 +220,32 @@ create_socket(char *name)
 	return s;
 }
 
+/* Update the modes on the socket. */
+static void
+update_socket_modes(int exec)
+{
+	struct stat st;
+	mode_t newmode;
+
+	if (stat(sockname, &st) < 0)
+		return;
+
+	if (exec)
+		newmode = st.st_mode | S_IXUSR;
+	else
+		newmode = st.st_mode & ~S_IXUSR;
+
+	if (st.st_mode != newmode)
+		chmod(sockname, newmode);
+}
+
 /* Process activity on the pty - Input and terminal changes are sent out to
 ** the attached clients. If the pty goes away, we die. */
 static void
 pty_activity(int s)
 {
 	unsigned char buf[BUFSIZE];
-	int len;
+	ssize_t len;
 	struct client *p;
 	fd_set readfds, writefds;
 	int highest_fd, nclients;
@@ -268,7 +293,7 @@ top:
 	/* Send the data out to the clients. */
 	for (p = clients, nclients = 0; p; p = p->next)
 	{
-		int written;
+		ssize_t written;
 
 		if (!FD_ISSET(p->fd, &writefds))
 			continue;
@@ -276,7 +301,7 @@ top:
 		written = 0;
 		while (written < len)
 		{
-			int n = write(p->fd, buf + written, len - written);
+			ssize_t n = write(p->fd, buf + written, len - written);
 
 			if (n > 0)
 			{
@@ -330,7 +355,7 @@ control_activity(int s)
 static void
 client_activity(struct client *p)
 {
-	int len;
+	ssize_t len;
 	struct packet pkt;
 
 	/* Read the activity. */
@@ -351,7 +376,10 @@ client_activity(struct client *p)
 
 	/* Push out data to the program. */
 	if (pkt.type == MSG_PUSH)
-		write(the_pty.fd, pkt.u.buf, pkt.len);
+	{
+		if (pkt.len <= sizeof(pkt.u.buf))
+			write(the_pty.fd, pkt.u.buf, pkt.len);
+	}
 
 	/* Attach or detach from the program. */
 	else if (pkt.type == MSG_ATTACH)
@@ -412,6 +440,8 @@ master_process(int s, char **argv, int waitattach, int statusfd)
 	int highest_fd;
 	int nullfd;
 
+	int has_attached_client = 0;
+
 	/* Okay, disassociate ourselves from the original terminal, as we
 	** don't care what happens to it. */
 	setsid();
@@ -419,6 +449,19 @@ master_process(int s, char **argv, int waitattach, int statusfd)
 	/* Set a trap to unlink the socket when we die. */
 	atexit(unlink_socket);
 
+	/* Create a pty in which the process is running. */
+	signal(SIGCHLD, die);
+	if (init_pty(argv, statusfd) < 0)
+	{
+		if (statusfd != -1)
+			dup2(statusfd, 1);
+		if (errno == ENOENT)
+			printf("%s: Could not find a pty.\n", progname);
+		else
+			printf("%s: init_pty: %s\n", progname, strerror(errno));
+		exit(1);
+	}
+
 	/* Set up some signals. */
 	signal(SIGPIPE, SIG_IGN);
 	signal(SIGXFSZ, SIG_IGN);
@@ -427,17 +470,7 @@ master_process(int s, char **argv, int waitattach, int statusfd)
 	signal(SIGTTOU, SIG_IGN);
 	signal(SIGINT, die);
 	signal(SIGTERM, die);
-	signal(SIGCHLD, die);
 
-	/* Create a pty in which the process is running. */
-	if (init_pty(argv, statusfd) < 0)
-	{
-		if (statusfd != -1)
-			dup2(statusfd, 1);
-		printf("%s: init_pty: %s\n", progname, strerror(errno));
-		exit(1);
-	}
-	
 	/* Close statusfd, since we don't need it anymore. */
 	if (statusfd != -1)
 		close(statusfd);
@@ -454,6 +487,8 @@ master_process(int s, char **argv, int waitattach, int statusfd)
 	/* Loop forever. */
 	while (1)
 	{
+		int new_has_attached_client = 0;
+
 		/* Re-initialize the file descriptor set for select. */
 		FD_ZERO(&readfds);
 		FD_SET(s, &readfds);
@@ -480,6 +515,16 @@ master_process(int s, char **argv, int waitattach, int statusfd)
 			FD_SET(p->fd, &readfds);
 			if (p->fd > highest_fd)
 				highest_fd = p->fd;
+
+			if (p->attached)
+				new_has_attached_client = 1;
+		}
+
+		/* chmod the socket if necessary. */
+		if (has_attached_client != new_has_attached_client)
+		{
+			update_socket_modes(new_has_attached_client);
+			has_attached_client = new_has_attached_client;
 		}
 
 		/* Wait for something to happen. */
@@ -507,7 +552,7 @@ master_process(int s, char **argv, int waitattach, int statusfd)
 }
 
 int
-master_main(char **argv, int waitattach)
+master_main(char **argv, int waitattach, int dontfork)
 {
 	int fd[2] = {-1, -1};
 	int s;
@@ -519,6 +564,28 @@ master_main(char **argv, int waitattach)
 
 	/* Create the unix domain socket. */
 	s = create_socket(sockname);
+	if (s < 0 && errno == ENAMETOOLONG)
+	{
+		char *slash = strrchr(sockname, '/');
+
+		/* Try to shorten the socket's path name by using chdir. */
+		if (slash)
+		{
+			int dirfd = open(".", O_RDONLY);
+
+			if (dirfd >= 0)
+			{
+				*slash = '\0';
+				if (chdir(sockname) >= 0)
+				{
+					s = create_socket(slash + 1);
+					fchdir(dirfd);
+				}
+				*slash = '/';
+				close(dirfd);
+			}
+		}
+	}
 	if (s < 0)
 	{
 		printf("%s: %s: %s\n", progname, sockname, strerror(errno));
@@ -530,7 +597,16 @@ master_main(char **argv, int waitattach)
 
 	/* If FD_CLOEXEC works, create a pipe and use it to report any errors
 	** that occur while trying to execute the program. */
-	if (pipe(fd) >= 0)
+	if (dontfork)
+	{
+		fd[1] = dup(2);
+		if (fcntl(fd[1], F_SETFD, FD_CLOEXEC) < 0)
+		{
+			close(fd[1]);
+			fd[1] = -1;
+		}
+	}
+	else if (pipe(fd) >= 0)
 	{
 		if (fcntl(fd[0], F_SETFD, FD_CLOEXEC) < 0 ||
 		    fcntl(fd[1], F_SETFD, FD_CLOEXEC) < 0)
@@ -542,6 +618,12 @@ master_main(char **argv, int waitattach)
 	}
 #endif
 
+	if (dontfork)
+	{
+		master_process(s, argv, waitattach, fd[1]);
+		return 0;
+	}
+
 	/* Fork off so we can daemonize and such */
 	pid = fork();
 	if (pid < 0)
@@ -565,7 +647,7 @@ master_main(char **argv, int waitattach)
 	if (fd[0] != -1)
 	{
 		char buf[1024];
-		int len;
+		ssize_t len;
 
 		close(fd[1]);
 		len = read(fd[0], buf, sizeof(buf));
@@ -594,6 +676,18 @@ openpty(int *amaster, int *aslave, char *name, struct termios *termp,
 	int master, slave;
 	char *buf;
 
+#ifdef _AIX
+	master = open("/dev/ptc", O_RDWR|O_NOCTTY);
+	if (master < 0)
+		return -1;
+	buf = ttyname(master);
+	if (!buf)
+		return -1;
+
+	slave = open(buf, O_RDWR|O_NOCTTY);
+	if (slave < 0)
+		return -1;
+#else
 	master = open("/dev/ptmx", O_RDWR);
 	if (master < 0)
 		return -1;
@@ -614,6 +708,7 @@ openpty(int *amaster, int *aslave, char *name, struct termios *termp,
 		return -1;
 	if (ioctl(slave, I_PUSH, "ldterm") < 0)
 		return -1;
+#endif
 #endif
 
 	*amaster = master;
@@ -658,6 +753,27 @@ forkpty(int *amaster, char *name, struct termios *termp,
 		buf = NULL;
 		if (ioctl(slave, TIOCSCTTY, NULL) < 0)
 			_exit(1);
+#elif defined(_AIX)
+		fd = open("/dev/tty", O_RDWR|O_NOCTTY);
+		if (fd >= 0)
+		{
+			ioctl(fd, TIOCNOTTY, NULL);
+			close(fd);
+		}
+
+		buf = ttyname(master);
+		fd = open(buf, O_RDWR);
+		close(fd);
+
+		fd = open("/dev/tty", O_WRONLY);
+		if (fd < 0)
+			_exit(1);
+		close(fd);
+
+		if (termp && tcsetattr(slave, TCSAFLUSH, termp) == -1)
+			_exit(1);
+		if (ioctl(slave, TIOCSWINSZ, winp) == -1)
+			_exit(1);
 #else
 		buf = ptsname(master);
 		fd = open(buf, O_RDWR);
